import re
'''
Types to arguments
'''


def range_int(left=None, right=None):

    def func(x):
        x = int(x)
        if left != None:
            x = max(left, x)
        if right != None:
            x = min(right, x)

        return x
    return func


def is_chan(chan):
    if chan and chan[0] != '#':
        raise TypeError('Is not a channel')
    return chan


def regexp(reg):
    compiled_reg = re.compile(reg)
    def func(x):
        matches = compiled_reg.match(x)
        if matches is None:
            raise TypeError('value is not corrected')
        return matches

    return func


from sqlalchemy.ext.declarative.api import declarative_base
from sqlalchemy.sql.functions import func
from sqlalchemy.sql.schema import Column, UniqueConstraint
from sqlalchemy.sql.sqltypes import Integer, DateTime, String

Base = declarative_base()


class Message(Base):
    __tablename__ = 'messages'
    id = Column(Integer, primary_key=True, autoincrement=True)
    created = Column(DateTime, default=func.now())
    channel = Column(String(30), nullable=False, index=True)
    key = Column(String(20), nullable=False, index=True)
    message = Column(String(200), nullable=False)



from grazyna.utils import register


@register(cmd='pep')
def utf(bot, number: int):
    """show pep X link"""
    msg = 'https://www.python.org/dev/peps/pep-{:04}/'.format(number)
    bot.reply(msg)



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, Session as BaseSession
from sqlalchemy.engine.reflection import Inspector

from .models import Base
from contextlib import contextmanager


class Session(BaseSession):

    @contextmanager
    def scope(self):
        "from http://docs.sqlalchemy.org/en/rel_1_0/orm/session_basics.html"
        try:
            yield self
            self.commit()
        except:
            self.rollback()
            raise
        finally:
            self.close()


def get_engine(uri):
    engine = create_engine(uri)
    inspector = Inspector.from_engine(engine)
    if not inspector.get_table_names():
        create_database(engine)
    return engine


def get_session(engine):
    return sessionmaker(
        bind=engine,
        class_=Session,
        expire_on_commit=False
    )()


def create_database(engine):
    Base.metadata.create_all(engine)



#!/usr/bin/python3

from grazyna.utils import create_help, register
from grazyna.utils.types import range_int
from grazyna import format

from math import isnan
from random import random, randint
from collections import namedtuple
from functools import wraps

import operator
import math

special_vars = {
    'pi': math.pi,
    'e': math.e,
}

def reverse_func(func):
    @wraps(func)
    def inner(*args):
        return func(*args[::-1])
    return inner


operator_funcs = {
    #'op': (num, func),
    '+': (2, operator.add),
    '-': (2, operator.sub),
    '/': (2, operator.truediv),
    '//': (2, operator.floordiv),
    '%': (2, operator.mod),
    '*': (2, operator.mul),
    '**': (2, operator.pow),

    '=': (2, operator.eq),
    '>': (2, operator.gt),
    '<': (2, operator.lt),

    '&': (2, lambda a, b: bool(a) and bool(b)),
    '|': (2, lambda a, b: bool(a) or bool(b)),
    '^': (2, lambda a, b: bool(a) ^ bool(b)),

    '!': (1, lambda a: not bool(a)),
    '~': (1, lambda a: -a),

    '~-': (2, reverse_func(operator.sub)),
    '~/': (2, reverse_func(operator.truediv)),
    '~//': (2, reverse_func(operator.floordiv)),
    '~%': (2, reverse_func(operator.mod)),
    '~**': (2, reverse_func(operator.pow)),

    'sin': (1, math.sin),
    'cos': (1, math.cos),
    'log': (1, math.log),
    'tan': (1, math.tan),
    'log10': (1, math.log10),
    'log2': (1, math.log2),

    '?': (0, random),
    '??': (2, lambda a, b: randint(int(a), int(b))),
}


@register(cmd="onp")
def onp(bot, tasks, round:range_int(0)=3):
    """
        example: .onp 2 2 2 + * log2 round=5
        operators https://github.com/firemark/grazyna/blob/master/grazyna/plugins/onp.py
    """
    score = calc(tasks.split(), round)
    bot.say("Result: %s" % score)


@register(cmd="rpn")
def rpn(bot, tasks, round:range_int(0)=3):
    yield from onp(bot, tasks, round)


def calc(tasks, round_num=3):
    buffer = []
    for task in tasks:
        num = get_number(task)
        if num is None:
            result = execute(task, buffer)
            if result is not None:
                return result
        else:
            buffer.append(num)

    if len(buffer) > 0:
        value = buffer.pop()
        if type(value) == bool:
            return value.__repr__()
        else:
            if isnan(value):
                if random() > 0.87:
                    return "NaNNaNNaNNaNNaN Batman!"
            return round(value, round_num)
    else:
        return "WTF"


def get_number(task):
    try:
        return float(task)
    except ValueError:
        return special_vars.get(task)


def execute(op, buffer):
    try:
        len_args, func = operator_funcs[op]
    except KeyError:
        return 'WTF - %s is not a good operator' % op

    if len_args > 0:
        args = buffer[-len_args:]
        if len(args) != len_args:
            return "WTF"
        del buffer[-len_args:]
    else:
        args = []

    try:
        value = func(*args)
    except OverflowError:
        return format.bold("Over 9000!!")
    else:
        buffer.append(value)




import asyncio
from functools import wraps
from datetime import datetime
from asyncio import async

from .exc import NoSuchNickError
from .models import User

import os.path


def whois_command(func):

    @wraps(func)
    def inner(self):
        future = self.get_from_whois()
        if future:
            return func(self, future.data)

    return inner


class MessageController(object):

    NUMERIC_COMMANDS = {
        '005': 'start',
        '330': 'whois_account',  # I don't have idea where is in RFC
        '311': 'whois_user',
        '312': 'whois_server',
        '313': 'whois_operator',
        '319': 'whois_channels',
        '318': 'whois_end',
        '401': 'no_such_nick',
    }
    log_files = {}

    __slots__ = ('user', 'data', 'protocol', 'command', 'prefix')

    def __init__(self, protocol, data):
        self.protocol = protocol
        if not data:
            self.data = []
            return
        prefix_or_not = data[0]

        if prefix_or_not[0] == ':':
            self.prefix = prefix_or_not[1:]  # without ':'
            self.user = User(self.prefix)
            self.command = data[1]
        else:
            self.command = data[0]

        self.data = data

    @property
    def config(self):
        return self.protocol.config

    def execute_message(self):
        if not self.data:
            return
        command = self.command.lower()
        command = self.NUMERIC_COMMANDS.get(command, command)
        method = getattr(self, 'command_%s' % command, None)

        if method is not None:
            method()

    def get_from_whois(self, pop=False):
        nick = self.data[3]
        heap = self.protocol.whois_heap
        return heap.get(nick) if not pop else heap.pop(nick, None)

    def log(self, channel, msg):
        str_datetime = datetime.now().strftime("%d-%m-%Y %H:%M:%S")
        if channel not in self.log_files:
            self.log_files[channel] = open(
                os.path.join(self.config.get('main', 'dir_log'), channel), "at"
            )
        file_log = self.log_files[channel]
        file_log.write('[%s] %s\r\n"' % (str_datetime, msg))
        file_log.flush()

    def command_join(self):
        channel = self.data[2]
        self.log(channel, '%s(%s)::JOIN' % (self.user.nick, self.user.prefix))

    def command_ping(self):
        self.protocol.send('PONG', self.data[1])

    def command_pong(self):
        self.protocol.ping_counter = 0

    def command_part(self):
        channel = self.data[2]
        if len(self.data) == 4:
            reason = self.data[3]
            self.log(channel, "%s::PART[%s]" % (self.user.nick, reason))
        else:
            self.log(channel, "%s::PART" % self.user.nick)

    def command_kick(self):
        channel, nick = self.data[2:4]
        self.log(channel, "%s KICK %s" % (self.user.nick, nick))

        if nick == self.protocol.config['main']['nick']:
            self.protocol.send('JOIN', channel)

    def command_error(self):
        pass

    def command_privmsg(self):
        channel, text = self.data[2:]
        self.log(channel, "<%s> %s" % (self.user.nick, text))
        async(self.protocol.importer.execute(channel, self.user, text))

    def command_notice(self):
        self.command_privmsg()

    def command_start(self):
        if self.protocol.ready:
            return

        self.protocol.ready = True
        async(ping_pong(self.protocol))

        for channel in self.protocol.config.getlist('main', 'channels'):
            self.protocol.send('JOIN', channel)

        kwargs = dict(self.protocol.config.items('auth'))
        del kwargs['module']
        cls_auther = self.protocol.config.getmodule('auth', 'module')
        cls_auther(**kwargs).auth(self.protocol)

    @whois_command
    def command_whois_account(self, whois_data):
        whois_data.account = self.data[4]

    @whois_command
    def command_whois_user(self, whois_data):
        nick, realname, host, _, ircname = self.data[3:8]
        whois_data.nick = nick
        whois_data.host = host
        whois_data.realname = realname
        whois_data.ircname = ircname

    @whois_command
    def command_whois_server(self, whois_data):
        whois_data.server = self.data[2]

    @whois_command
    def command_whois_channels(self, whois_data):
        whois_data.channels = [
            channel.strip('@%+') for channel in self.data[2].split()
        ]

    def command_whois_end(self):
        future = self.get_from_whois(pop=True)
        if future:
            future.set_result(future.data)

    def command_no_nick(self):
        future = self.get_from_whois(pop=True)
        if future:
            future.set_exception(NoSuchNickError)


@asyncio.coroutine
def ping_pong(protocol):
    while True:
        protocol.ping_counter += 1
        protocol.send_msg('PING', '1337')
        yield from asyncio.sleep(64)
        if protocol.ping_counter > 4:  # 4 * 64 = 256 seconds
            break
    protocol.transport.close()



from grazyna.utils import register

from urllib.parse import quote


BASE_URL = 'http://lmgtfy.com/?q='


@register(cmd='lmgtfy')
def lmgtfy(bot, query):
    """let me google that for you"""

    url = BASE_URL + quote(query)
    bot.say(url)



from grazyna.models import Message
from grazyna.utils import register


@register(cmd='add', on_private=False)
def add_cmd(bot, key, message):
    if not bot.protocol.db:
        return
    with bot.protocol.get_session() as session:
        session.add(Message(
            key=key,
            channel=bot.chan,
            message=message,
        ))
    bot.reply('Done!')


@register(cmd='del', on_private=False, admin_required=True)
def delete_cmd(bot, key):
    if not bot.protocol.db:
        return
    with bot.protocol.get_session() as session:
        session.query(Message).filter_by(key=key, channel=bot.chan).delete()
    bot.reply('Done!')




'''
Created on 31-03-2013

@author: firemark
'''
from asyncio import coroutine


class RequestBot(object):
    """First argument in register functions.
    Has useful informations as channel, nick
    and methods to communicate with user/channel"""

    __slots__ = ('private', 'user', 'chan', 'protocol', 'config', 'plugin',
                 'temp')

    def __init__(self, protocol, *, user=None, chan=None, private=False,
                 config=None, plugin=None, temp=None):
        self.protocol = protocol
        self.private = private
        self.user = user
        self.chan = chan
        self.config = config
        self.plugin = plugin
        self.temp = temp

    @coroutine
    def is_admin(self, nick=None):
        nick = nick or self.user.nick
        result = yield from self.protocol.whois(nick)
        return result.account in self.protocol.config.getlist('main', 'admins')

    @property
    def nick_chan(self):
        return self.user.nick if self.private else self.chan

    def say(self, msg, nick_chan=None):
        nick_chan = nick_chan or self.nick_chan
        self.protocol.say(nick_chan, msg)

    def notice(self, msg, nick=None):
        self.protocol.notice(nick or self.user.nick, msg)

    def reply(self, msg, chan=None, nick=None):
        self.protocol.reply(nick or self.user.nick, msg, chan or self.chan)

    def kick(self, who=None, why='', chan=None):
        if not self.private:
            self.protocol.kick(who or self.user.nick, chan or self.chan, why)

    def private_say(self, msg, nick=None):
        self.protocol.say(nick or self.user.nick, msg)

    def command(self, *args):
        self.protocol.send(*args)

    def command_msg(self, *args):
        self.protocol.send_msg(*args)

    def mode(self, flag, arg, chan=None):
        self.protocol.mode(chan or self.chan, flag, arg)

    def time_ban(self, time, who=None, why='', prefix=None, chan=None):
        chan = chan or self.chan
        prefix = prefix or "*!%s@%s" % (self.user.realname, self.user.host)
        who = who or self.user.nick
        self.protocol.time_ban(time, who, chan, why, prefix)



from . import format

from asyncio import async, coroutine
from asyncio.futures import Future
from grazyna.models import Message

from .request import RequestBot
from importlib import reload
from collections import defaultdict
from inspect import getcallargs
from datetime import datetime, timedelta
from sqlalchemy.sql.functions import func

import re
import traceback
import sys

re_split = re.compile(r' *(?:(\w+)= *)?(?:"([^"]+)"|(\S+))')  # parse args
re_db_args = re.compile(r'(^|[^$])\$(@|\d+)')


class Plugin(list):

    __slots__ = ('temp', 'name', 'module_path', 'future', 'config')

    def __init__(self, name, module_path, funcs):
        super().__init__(funcs)
        self.temp = {}
        self.name = name
        self.module_path = module_path
        self.config = {}
        self.future = Future()

    def __repr__(self):
        return "Plugin('%s')" % self.name


class ExecutedCounter(object):

    __slots__ = ('last_time', 'counter')

    def __init__(self):
        self.last_time = datetime.now()
        self.counter = 0

    def inc(self):
        self.counter += 1


class ModuleManager(object):

    protocol = None
    plugins = None
    executed_counters = None

    def __init__(self, protocol):
        self.protocol = protocol
        self.plugins = {}
        self.executed_counters = defaultdict(ExecutedCounter)

    @property
    def config(self):
        return self.protocol.config

    def load_all(self):
        for name, module_path in self.config.items('plugins'):
            self.load(name, module_path)

    def cancel_tasks(self):
        for plugin in self.plugins.values():
            plugin.future.cancel()

    def load(self, name, module_path):
        path, module_name = module_path.rsplit(".", 1)
        module = __import__(module_path, globals(), locals(), [module_name])
        reload(module)
        self.plugins[name] = plugin = Plugin(
            name, module_path,
            [
                obj for obj in module.__dict__.values()
                if getattr(obj, 'is_bot_event', False) is True
            ]
        )

        cfg = self.config
        plugin_name = 'plugin:%s' % name
        if cfg.has_section(plugin_name):
            plugin.config.update(cfg.items(plugin_name))

        initials_funcs = (
            func for func in module.__dict__.values()
            if getattr(func, 'is_initial', False) is True
        )

        for initial_func in initials_funcs:
            initial_func(plugin, self.protocol)

        future = plugin.future
        loops = [
            func(self.protocol, plugin, future)
            for func in module.__dict__.values()
            if getattr(func, 'is_loop', False) is True
        ]

        for loop in loops:
            async(loop)

    def reload(self, name):
        plugin = self.plugins[name]
        plugin.future.cancel()
        self.load(name, plugin.module_path)

    def remove(self, name):
        plugin = self.plugins.pop(name)
        plugin.future.cancel()

    @coroutine
    def execute(self, channel, user, msg):
        cmd, text = self.get_cmd_and_text(msg)
        private = channel[0] != "#"

        if cmd is not None:
            yield from self.execute_command(cmd, text, private, channel, user)
        else:
            yield from self.execute_regs(msg, private, channel, user)

    @coroutine
    def execute_command(self, cmd, text, private, channel, user):
        plugin, func = self.find_command(cmd, private, channel)
        if func is None:
            if private:
                return
            yield from self.find_message_in_db(cmd, channel, text)
            return

        args, kwargs = get_args_from_text(text, func.max_args)
        yield from self.execute_func(
            func, plugin, private, channel, user, args, kwargs)


    @coroutine
    def find_message_in_db(self, cmd, channel, text):
        if not self.protocol.db:
            return
        with self.protocol.get_session() as session:
            row = (
                session
                    .query(Message.message)
                    .filter_by(key=cmd, channel=channel)
                    .order_by(func.random())
                    .first()
            )
        if row is None:
            return

        msg = row.message
        max_args = max(
            (
                int(m) + 1 if m.isnumeric() else 0
                for m in (m.group(2) for m in re_db_args.finditer(msg))
            ), default=0
        )
        args, kwargs = get_args_from_text(text, max_args)

        def replace_dollars(match):
            val = match.group(2)
            if val == '@':
                return text
            return match.group(1) + args[int(val)]

        try:
            msg = re_db_args.sub(replace_dollars, msg)
        except IndexError:
            return

        return self.protocol.say(channel, msg)

    @coroutine
    def execute_regs(self, msg, private, channel, user):
        for plugin, func, matches in self.find_regs(msg, private):
            for match, _ in zip(matches, range(3)):
                args = match.groups()
                kwargs = match.groupdict()
                yield from self.execute_func(
                    func, plugin, private, channel, user, args, kwargs)

    def find_function_in_plugin_with_name(self, private=None):
        return (
            (plugin, func)
            for name, plugin in self.plugins.items()
            for func in plugin
            if private is None
            or (not private and func.on_channel)
            or (private and func.on_private)
        )

    def find_command(self, cmd, private=None, channel=None):
        return next(
            (
                (plugin, func) for plugin, func
                in self.get_commands(private)
                if self.cmd_is_good(plugin, func, cmd, private, channel)
            ),
            (None, None)
        )

    def cmd_is_good(self, plugin, func, cmd=None, private=False, channel=None):
        cfg = self.get_plugin_cfg(plugin.name)

        if cmd is not None and func.cmd.format(**cfg) != cmd:
            return False

        if private or channel is None:
            return True

        def get_list(key):
            return [
                x for x in (x.strip() for x in cfg.get(key, '').split(','))
                if x
            ]

        whitelist = get_list('whitelist')
        if whitelist:
            return channel in whitelist

        blacklist = get_list('blacklist')
        return channel not in blacklist


    def get_commands(self, private=None):
        return (
            (plugin, func) for plugin, func
            in self.find_function_in_plugin_with_name(private)
            if func.is_reg is False
        )

    def find_regs(self, msg, private):
        return (
            (plugin, func, func.compiled_reg.finditer(msg))
            for plugin, func in self.find_function_in_plugin_with_name(private)
            if func.is_reg is True
        )

    def get_plugin_cfg(self, name):
        cfg = self.config
        conf = self.plugins[name].config.copy()
        conf.update(
            __nick__=cfg.get('main', 'nick')
        )

        return conf

    def get_cmd_and_text(self, msg):
        cfg = self.config['main']
        startswith = lambda key: msg.startswith(cfg[key])
        if startswith('command-prefix'):
            data = msg.split(maxsplit=1)
            cmd = data[0][1:]
        elif startswith('nick'):
            data = msg.split(maxsplit=2)[1:]
            cmd = data[0]
        else:
            return None, msg
        text = data[1] if len(data) > 1 else ""
        return cmd, text

    @coroutine
    def execute_func(self, func, plugin, private, channel, user, args, kwargs):
        if func.block and self.is_blocked(user):
            return  # filtr antyspamowy, nienawidze was

        bot = RequestBot(
            protocol=self.protocol,
            private=private,
            chan=channel if not private else None,
            config=self.get_plugin_cfg(plugin.name),
            plugin=plugin,
            user=user,
            temp=plugin.temp
        )

        try:
            dict_arg = check_type(args, kwargs, func)
        except Exception as e:
            print(type(e), e)
            return

        yield from self._execute_func(func, bot, dict_arg, user, channel)

    @coroutine
    def _execute_func(self, func, bot, args, user, channel):
        if func.admin_required:
            is_admin = yield from bot.is_admin()
            if not is_admin:
                return
        try:
            yield from func(bot, **args)
        except Exception:
            traceback.print_exc(file=sys.stdout)
            tb = traceback.format_exc().split('\n')[-4:-1]
            self.protocol.reply(
                user.nick,
                '{} {} {}'.format(
                    format.bold('ERR:'),
                    tb[2],
                    format.color(tb[0], format.color.red),
                ),
                channel
            )

    def is_blocked(self, user):
        now = datetime.now()
        delta = timedelta(
            seconds=self.config.getint('main', 'time_to_block')
        )

        self.executed_counters = defaultdict(ExecutedCounter, { # cleanups
            key: counter for key, counter in self.executed_counters.items()
            if now - counter.last_time < delta
        })

        nick = user.prefix
        counter = self.executed_counters[nick]

        executed_commands_per_time = self.config.getint(
            'main', 'executed_commands_per_time'
        )

        if counter.counter > executed_commands_per_time:
            return True
        else:
            counter.inc()
            return False


def get_args_from_text(text, max_args=-1):
    if not text:
        return [], {}
    args = []
    kwargs = {}
    for arg_group in re_split.finditer(text):
        arg_name, arg_quotes, arg = arg_group.groups()
        arg = arg_quotes or arg

        if arg_name:
            kwargs[arg_name] = arg
        else:
            args.append(arg)

    if len(args) > max_args and max_args != -1:
        last_arg = ' '.join(args[max_args:])
        args = args[0:max_args]
        try:
            args[-1] += ' ' + last_arg
        except IndexError:
            args = [last_arg]

    return args, kwargs


def check_type(args, kwargs, func):
    real_func = getattr(func, '__wrapped__', func)
    dict_arg = getcallargs(real_func, None, *args, **kwargs)
    for key, item in dict_arg.items():
        arg_type = real_func.__annotations__.get(key, str)
        if item is not None:
            dict_arg[key] = arg_type(item)
    del dict_arg['bot']
    return dict_arg
